using System.Collections;
using SpellBoundAR.QRCodes;
using Unity.Collections;
using UnityEngine;
using UnityEngine.XR.ARFoundation;
using UnityEngine.XR.ARSubsystems;
using ZXing;

namespace SpellBoundAR.ARImageTracking
{
    public class ARFoundationQRCodeScanner : QRCodeScanner
    {
        [Header("References")]
        [SerializeField] private ARCameraManager arCameraManager;

        [Header("Cache")]
        private bool _isProcessing;
        private XRCpuImage _image;
        private NativeArray<byte> _rawData;
        
        private bool _cachedDimensions;
        private int _cachedWidth;
        private int _cachedHeight;

        public override int Width => _cachedDimensions ? _cachedWidth : Screen.width;
        public override int Height => _cachedDimensions ? _cachedHeight : Screen.height;

        public override void Scan()
        {
            if (_isProcessing) return;
            if (!arCameraManager.TryAcquireLatestCpuImage(out _image)) return;
            
            _isProcessing = true;
            _cachedDimensions = true;
            _cachedWidth = _image.width;
            _cachedHeight = _image.height;
            StartCoroutine(ProcessImage(_image));
            _image.Dispose();
        }

        IEnumerator ProcessImage(XRCpuImage image)
        {
            XRCpuImage.AsyncConversion request = image.ConvertAsync(new XRCpuImage.ConversionParams
            {
                inputRect = new RectInt(0, 0, image.width, image.height),
                outputDimensions = new Vector2Int(image.width, image.height),
                outputFormat = TextureFormat.R8,
                transformation = XRCpuImage.Transformation.MirrorY
            });

            while (!request.status.IsDone())
                yield return null;

            if (request.status != XRCpuImage.AsyncConversionStatus.Ready)
            {
                Debug.LogWarningFormat("ProcessImage: Request failed with status {0}", request.status);
                request.Dispose();
                _isProcessing = false;
                yield break;
            }

            _rawData = request.GetData<byte>();
            byte[] grayscalePixels = _rawData.ToRawBytes();

            CreateDebugTexture();
            UpdateDebugTexture();
            
            StartCoroutine(Decode(grayscalePixels, image.width, image.height, RGBLuminanceSource.BitmapFormat.Gray8));
            
            request.Dispose();
            _isProcessing = false;
        }
        
        public override void CreateDebugTexture()
        {
            if (!showDebugTexture || DebugTexture || !_cachedDimensions) return;
            DebugTexture = new Texture2D(
                _cachedWidth,
                _cachedHeight,
                TextureFormat.R8,
                false);
        }

        public override void UpdateDebugTexture()
        {
            if (!showDebugTexture || !DebugTexture) return;
            DebugTexture.LoadRawTextureData(_rawData);
            DebugTexture.Apply();
        }
    }

    /// <summary>
    /// https://gist.github.com/asus4/992ae563ca1263b1dd936b970e7fc206
    /// </summary>
    public static class NativeArrayExtensions
    {
        public static byte[] ToRawBytes<T>(this NativeArray<T> arr) where T : struct
        {
            var slice = new NativeSlice<T>(arr).SliceConvert<byte>();
            var bytes = new byte[slice.Length];
            slice.CopyTo(bytes);
            return bytes;
        }
    }
}
